package agenda;

import schema.item.Items;
import element.IntElement;
import element.StringPosition;

/**
 * Priority queue agenda implementation where the priority of items
 * is determined by an heuristic function based on the last two elements of the item, which are considered
 * to be the j-index (progress, string position) and the distance (integer).
 * 
 * When the last two elements do not match these types, the highest priority (0) is returned.
 * 
 * @author carlos
 *
 * Created at regulus, 11-nov-2006 13:25:19
 */
public class HeuristicPriorityQueueAgenda extends PriorityQueueAgenda
{
	
	//heuristic is:   progress - k * distance
	
	//in order to be able to use non-integer k's to some point, we make k = kNumer / kDenom, so that
	//heuristic is:   progress - (kNumer/kDenom) * distance
	//and the priority is an integer:  kDenom * progress - kNumer * distance
	
	private int K_DENOMINATOR = 1;
	private int K_NUMERATOR = 10;
	
	//configurable progress function: progress = c1 * i + c2 * j
	
	private double PROGRESS_COEF_I = -0.5;
	private double PROGRESS_COEF_J = 1;
	
	
	/**
	 * Obtains the priority of an item.
	 * @param item
	 */
	protected int getPriority ( Object item )
	{
		Object[] theItem = (Object[]) item;
		Object lastElement = theItem[theItem.length-1];
		Object beforeLastElement = theItem[theItem.length-2];
		if ( theItem.length < 3 )
			return 0;
		Object beforeBeforeLastElement = theItem[theItem.length-3];
		if ( lastElement instanceof IntElement && !( lastElement instanceof StringPosition )
				&&  beforeLastElement instanceof StringPosition  )
		{
			int j = ( (IntElement) beforeLastElement ).getValue();
			int i = ( (IntElement) beforeBeforeLastElement ).getValue();
			int progress = (int)(PROGRESS_COEF_I * i) + (int)(PROGRESS_COEF_J * j);
			int distance = ( (IntElement) lastElement ).getValue();
			int sentLength = StringPosition.getMaxValue(); //equals max progress
			
			int heuristic = K_DENOMINATOR * progress - K_NUMERATOR * distance;
			
			int priority = sentLength - heuristic;
			
			if ( priority < 0 ) 
			{
				priority = 0;
				System.err.println("[WARNING] Less than zero priority found (HeuristicPriorityQueueAgenda). Priority has been adjusted to zero.");
				System.err.println("          Offending item was: " + Items.toString(theItem));
				System.err.println("          heu: " + heuristic + " sentlen: " + sentLength);
			}
			
			//System.out.println(priority);
			
			return priority;
		}
		return 0;
	}
	
}