package element;

/**
 * @author carlos
 * 
 * Clase que representa un smbolo gramatical, terminal o no.
 */
public class Symbol extends SimpleElement implements HashIndexable
{
	
	/**
	 * Nmero que identifica a este smbolo.
	 * Slo tiene sentido para smbolos ground.
	 */
	private int id;
	
	/**
	 * Atributo que nos dice si el smbolo es constante (ground) o no.
	 */
	private boolean ground;
	
	/**
	 * Crea un smbolo con el nmero identificador dado, instanciado o no segn
	 * el valor del parmetro ground.
	 */
	public Symbol ( int id , boolean ground )
	{
		this.id = id; this.ground = ground;
	}
	
	/**
	 * Nos dice si este smbolo es ground (constante/instanciado) o no.
	 */
	public boolean isGround()
	{
		return ground;
	}
	
	/**
	 * Devuelve la ID de un smbolo ground dado.
	 * @return id de este smbolo, si es ground. Si no lo es, devuelve
	 * un valor indefinido.
	 */
	public final int getID()
	{
		return id;
	}
	
	/**
	 * Devuelve el nombre de un smbolo ground dado.
	 * @return nombre de este smbolo, si es ground. Si no lo es, devuelve
	 * null.
	 */
	public String getName()
	{
		return SymbolTable.instance().getName(id);
	}
	
	/**
	 * Nos dice si este smbolo empareja con el smbolo constante dado.
	 */
	public boolean matches ( Element e )
	{
		if ( !(e instanceof Symbol) ) return false;
		if ( !this.isGround() ) return true;
		else return id == ((Symbol)e).getID();
	}
	
	/**
	 * Nos da una representacin legible del smbolo.
	 */
	public String toString()
	{
		if ( ground )
			return SymbolTable.instance().getName(this);
		else
			return "_" + SymbolTable.instance().getName(this);
	}
	
	/**
	 * Mtodo de parsing para smbolos, nos da el smbolo asociado a una cadena dada.
	 * Ntese que se comparten smbolos, i.e. no siempre se construye uno nuevo.
	 * @return Smbolo correspondiente a la cadena dada.
	 */
	public static Symbol fromString(String s)
	{
		return SymbolTable.instance().symbolFor(s);
	}
	
	/**
	 * Obtiene un smbolo no-ground para una cadena dada.
	 * Importante: no hacer esto para cadenas correspondientes a smbolos ground.
	 * Los conjuntos de nombres de smbolos ground y no-ground
	 * no deben solaparse.
	 * @param s Nombre del smbolo no-ground.
	 * @return Smbolo no-ground correspondiente a la cadena dada.
	 */
	public static Symbol nonGroundFromString ( String s )
	{
	    Symbol sy = SymbolTable.instance().symbolFor(s);
	    sy.setGround(false);
	    return sy;
	}
	
	/**
	 * Hace que el smbolo sea, o no, ground (instanciado).
	 * @param b Si el smbolo es ground o no.
	 */
	public void setGround ( boolean b )
	{
	    this.ground = b;
	}
	
	/**
	 * De ArrayIndexable. Devuelve la ID numrica del smbolo, como getID().
	 * @see element.ArrayIndexable#getArrayLength()
	 */
    public int getArrayIndex ()
    {
        return id;
    }

    /**
     * De ArrayIndexable. Ntese que este mtodo slo devolver el valor correcto segn
     * lo definido en el interfaz cuando todos los smbolos que se vayan a utilizar
     * en el parsing hayan sido creados.
     * @see ArrayIndexable
     */
    public static int getArrayLength ()
    {
        return SymbolTable.instance().getLastId() + 1;
    }
        
    public Object getHashKey()
    {
        return this;
    }
	
    public String getExpressionCode()
    {
    	return "element.SymbolTable.instance().symbolFor(\""+SymbolTable.instance().getName(this)+"\")";
    }
	
}