package engine;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Date;
import java.util.List;

import schema.step.DeductiveStep;
import schema.item.ItemHandler;
import schema.item.Items;
import schema.step.StepHandler;
import java.util.List;
import element.IntElement;

import agenda.Agenda;

/**
 * Variation of the standard deductive engine where the parsing process finishes as soon as a final item is found.
 * 
 * @author carlos
 *
 * Created at regulus, 30-abr-2007 19:44:14
 */
public class FirstSolutionDeductiveEngine implements DeductiveEngine
{
	
	protected Agenda agenda;
	protected List deductiveSteps;
	protected List goalConditions;
	protected ItemHandler itemHandler;
	protected StepHandler stepHandler;
	
	public void setItemHandler ( ItemHandler ih )
	{
		itemHandler = ih;
	}
	
	public FirstSolutionDeductiveEngine ( Agenda agenda , List deductiveSteps , List goalConditions , ItemHandler itemHandler , StepHandler stepHandler )
	{
		this.agenda = agenda;
		this.deductiveSteps = deductiveSteps;
		this.goalConditions = goalConditions;
		this.itemHandler = itemHandler;
		this.stepHandler = stepHandler;
		EngineManager.register(this);
	}
	
	
	private boolean preprocessing = false;
	
	private void emptyAgenda()
	{
		agenda.clear();
	}
	
	/**
	 * Generates all the items that can be deducted from the items in the agenda.
	 * In this concrete class, process() is the same as preprocess().
	 * This deductive engine just triggers all possible deductions with the items in the agenda, without any further consideration.
	 * This is valid both for preprocessing when only the grammar is known (and steps requiring no knowledge of the string, such as
	 * Earley Initter, are executed) and for executing the actual algorithm.
	 * @return List of all the items in the ItemHandler.
	 */
	public List preprocess ()
	{
		dumpStatsHeader();
		resetStatistics();
		preprocessing = true;
		List l = process ();
		preprocessing = false;
		return l;
	}
		
	/*statistics*/
	int processedItems = 0; //continually updated
	
	//updated only on end process:
	int generatedItems = 0;
	int leftInAgenda = 0;
	long processTime = 0;
	String finalItem = "";
	
	//not reset
	int sentenceNumber = 0;
	
	private void resetStatistics()
	{
		processedItems = 0;
		generatedItems = 0;
		leftInAgenda = 0;
		finalItem = "";
	}
	
	
	private void processEmptyAntecedentSteps()
	{
		
		//Aplicar pasos de antecedente vaco y quitarlos de la lista de pasos aplicables.
		int sz = deductiveSteps.size();
		Object[] emptyItem = new Object[0];
		for ( int i = sz-1 ; i >= 0 ; i-- )
		{
			List l = ((DeductiveStep)deductiveSteps.get(i)).applyTo(emptyItem);
			if ( l!=null )
			{
				for ( int j = 0 ; j < l.size() ; j++ )
				{
						Object[] newItem = (Object[]) l.get(j);
						if ( !itemHandler.exists(newItem) )
						{
							itemHandler.addItem(newItem);
							agenda.add(newItem);
						}
				}
				if ( l.size() > 0 )
				{
					if ( ((DeductiveStep)deductiveSteps.get(i)).hasEmptyAntecedent() )
						deductiveSteps.remove(i);
				}
			}
		}
		
	}
	
	
	
	/**
	 * Returns whether the given item is final according to the goal conditions associated to the schema
	 * this deductive engine was instantiated with.
	 */
	private boolean isFinal( Object[] item )
	{

		if ( goalConditions == null || goalConditions.size() == 0 ) { return true; }
		else
		{
			for ( int j = 0 ; j < goalConditions.size() ; j++ )
			{
				List result = ((DeductiveStep)goalConditions.get(j)).applyTo(item);
				if ( result != null && result.size() > 0 )
				{
					return true;
				}
			}
			return false;
		}
		
	}
	
	
	
	/**
	 * Generates all the items that can be deducted from the items in the agenda.
	 * In this concrete class, process() is the same as preprocess().
	 * The standard deductive engine just triggers all possible deductions with the items in the agenda, without any further consideration.
	 * This is valid both for preprocessing when only the grammar is known (and steps requiring no knowledge of the string, such as
	 * Earley Initter, are executed) and for executing the actual algorithm.
	 * @return List of all the items in the ItemHandler.
	 */
	public List process()
	{
		
		resetStatistics();
		if ( !preprocessing )
			sentenceNumber++;
			
		processTime = System.currentTimeMillis();
		
		//Aplicar pasos de antecedente vaco y quitarlos de la lista de pasos aplicables.
		processEmptyAntecedentSteps();
		
		int sz = 0;
		//aplicar pasos con antecedente, usando agenda
		while ( !agenda.isEmpty() )
		{
			Object[] curItem = (Object[]) agenda.removeFirst();
			
			processedItems++;
			
			if ( processedItems % 10000 == 0 )
			{
				System.out.println("<"+processedItems+">" + " " + Items.toString(curItem));
			}
				
			List pasosDeductivosAplicables = stepHandler.getByItem(curItem);
			
			sz = pasosDeductivosAplicables.size();
			for ( int i = 0 ; i < sz ; i++ )
			{
				List l = ((DeductiveStep)pasosDeductivosAplicables.get(i)).applyTo(curItem);
				if ( l != null )
				{
					int sz2 = l.size();
					for ( int j = 0 ; j < sz2 ; j++ )
					{
						Object[] newItem = (Object[]) l.get(j);
						if ( !itemHandler.exists(newItem) )
						{
							if ( isFinal(newItem) && !preprocessing )
							{
								itemHandler.addItem(newItem);
								
								List lItems = (itemHandler.getAll());
								
								processTime = System.currentTimeMillis()-processTime;
								generatedItems = lItems.size();
								leftInAgenda = agenda.size();
								finalItem = schema.item.Items.toString(newItem);
								
								System.out.println("Final item found: " + finalItem);
								System.out.println("Items processed: " + processedItems);
								System.out.println("Items generated: " + generatedItems);
								System.out.println("Items left in agenda: " + leftInAgenda);
								dumpStatsLine();
								emptyAgenda();
								
								//dumpLine("End of Items" , "c:\\dumpits.txt");
								
								return lItems;
							}
						
							int errCount = 0;
							
							if ( preprocessing )
							{
								//Chapuz (if-cond): do not error-correct on preprocessing.
								if ( newItem[newItem.length-1] instanceof IntElement )
								{
									errCount = ( (IntElement) newItem[newItem.length-1] ).getValue();
								}
							}
							
							if ( !preprocessing || errCount == 0 )
							{
								
								itemHandler.addItem(newItem);
								agenda.add(newItem);
								//dumpLine("Added: " + schema.item.Items.toString(newItem) , "c:\\dumpits.txt");
						
							}
							
						
						
						}
					}
				}
			}
		}
				
		List lItems = (itemHandler.getAll());
		
		processTime = System.currentTimeMillis()-processTime;
		generatedItems = lItems.size();
		leftInAgenda = agenda.size();
		finalItem = "[NONE]";
		
		System.out.println("Final item found: " + finalItem);
		System.out.println("Items processed: " + processedItems);
		System.out.println("Items generated: " + generatedItems);
		System.out.println("Items left in agenda: " + leftInAgenda);
		dumpStatsLine();
		
		//dumpLine("End of Items" , "c:\\dumpits.txt");
		
		emptyAgenda();
		
		return lItems;
		
	}
	
	
	public Object getProperty ( String key )
	{
		return null;
	}
	
	private String format ( int length , String str )
	{
		StringBuffer sb = new StringBuffer("");
		int spaces = length - str.length();
		for ( int i = 0 ; i < spaces ; i++ )
			sb.append(" ");
		sb.append(str);
		return sb.toString();
	}
	
	//private PrintWriter pw;
	
	
	public void dumpLine ( String line )
	{
		dumpLine ( line , "c:\\dump.txt" );
	}
	
	public void dumpLine( String line , String filename )
	{
		PrintWriter pw = null;
		try
		{
			if ( pw == null )
				pw = new PrintWriter ( new OutputStreamWriter ( new FileOutputStream ( new File ( filename ) , true ) ) );
			pw.println(line);
			pw.close();
		}
		catch ( Throwable t )
		{
			System.err.println("Error writing dump to file " + filename + ": ");
			t.printStackTrace();
		}
	}
	
	public void dumpStatsLine()
	{
		String line = getStatsLine();
		dumpLine(line);
	}
	
	public void dumpStatsHeader()
	{
		String line = getStatsHeader();
		dumpLine(line);
	}
	
	private String getStatsLine()
	{
		StringBuffer sb = new StringBuffer("");
		sb.append(format(15,String.valueOf(sentenceNumber)));
		sb.append(format(15,String.valueOf(processedItems)));
		sb.append(format(15,String.valueOf(generatedItems)));
		sb.append(format(15,String.valueOf(processTime)));
		sb.append(finalItem);
		return sb.toString();
	}
	
	private String getStatsHeader()
	{
		StringBuffer sb = new StringBuffer("");
		sb.append(format(15,"Sentence #"));
		sb.append(format(15,"Processed"));
		sb.append(format(15,"Generated"));
		sb.append(format(15,"Runtime"));
		sb.append(finalItem);
		return sb.toString();
	}
	
	
	
	
	
	
	
}