/*
 * Created at regulus on 08-oct-2005 11:50:53
 * as file Grammar.java on package grammar
 */
package grammar;

import java.io.File;
import java.io.IOException;

import element.Element;

/**
 * @author carlos
 * 
 * This interface represents any kind of grammar. In the context of this system,
 * a grammar is an entity which:
 * 
 * - Can be read from a file, using an initFromFile(File) method. Implementations
 * of Grammar must allow the creation of instances via default constructor, and
 * their initialization via initFromFile.
 * - Contains a list of Element arrays (Element[]). Each of these Element arrays
 * can be accessed via the getEntity(int) method, and the size() method returns the number
 * of them.
 * - The Element arrays may contain simple, composite and sequence elements. They
 * may not contain expressions.
 * - The Element arrays represent the entities used to instantiate the deductive
 * steps in a schema for a particular grammar. In simple cases such as the default
 * implementation of context-free grammars, these entities are the grammar rules.
 * However, this needn't always be the case. For example, in tree-adjoining grammars,
 * rules are too complex and we prefer to break them down into sets of context-free
 * rules, which are used to instantiate the steps. In this case, these context-free
 * rules are the grammatical entities stored in the Element arrays, and the real
 * TAG rules are implicit.
 * - The initFromFile(File) method may perform any kind of preprocessing on the grammar
 * or compute additional information (left-corner relationship tables, etc.)
 * that will be usable by deductive steps on instantiation or execution, since 
 * obviously the grammar initialization is performed before any step is instantiated 
 * or executed.
 *
 * Created at regulus, 08-oct-2005 11:50:53
 */
public interface Grammar 
{

	/**
	 * Initializes a Grammar from a file.
	 * Each concrete subclass of Grammar will expect its own file format to
	 * construct the Grammar.
	 * @param f File containing the grammar data.
	 * @throws IOException If there is any I/O accessing the file.
	 * @throws IncorrectGrammarFormatException If the file can be read, but
	 * has an unexpected format an the grammar can't be initialized.
	 */
	public void initFromFile ( File f ) throws IOException , IncorrectGrammarFormatException;
	
	/**
	 * @param i Index of the entity that we wish to obtain.
	 * @return The ith entity in the grammar.
	 */
	public Element[] getEntity ( int i );
	
	/**
	 * @return The number of entities in the grammar.
	 */
	public int size();
	
	
}
