package test;

import element.Element;
import element.StringPosition;
import element.Symbol;
import grammar.Grammar;
import grammar.IncorrectGrammarFormatException;
import grammar.GrammarLoader;
import agenda.Agenda;
import agenda.AgendaLoader;
import engine.DeductiveEngine;
import engine.DeductiveEngineLoader;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.PrintStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Iterator;
import java.util.Date;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.JPopupMenu;
import javax.swing.JMenuItem;

import schema.item.ItemHandler;
import schema.item.Items;
import schema.step.DeductiveStep;
import schema.step.StepFactory;
import schema.step.StepHandler;
import element.SymbolTable;

/**
 * This class implements a CLI (Command-Line Interface) for generated
 * parsers.
 * @author carlos
 *
 * Created at regulus, 26-nov-2006 18:57:34
 */
public class ParserConsole
{

	abstract class Command
	{
		private String commandName;
		public Command ( String commandName )
		{
			this.commandName = commandName;
		}
		public String getCommandName()
		{
			return commandName;
		}
		public abstract void execute ( String arguments );
	}
	
	
	class ParseFileCommand extends Command
	{
		public ParseFileCommand()
		{
			super("PARSEFILE");
		}
		public void execute ( String arguments )
		{
			StringTokenizer st = new StringTokenizer ( arguments );
			String arg1 = st.nextToken().trim();
			String arg2 = st.nextToken().trim();
			parseFile ( arg1, arg2 );
		}
		public String toString()
		{
			return "PARSEFILE <file-to-parse> <file-to-dump-results-to>";
		}
	}
	
	class ParseSentenceCommand extends Command
	{
		public ParseSentenceCommand()
		{
			super("PARSE");
		}
		public void execute ( String arguments )
		{
			StringTokenizer st = new StringTokenizer(arguments," \n\r\t",true);
			boolean finalsOnly = false; //-f = finals only
			boolean noItems = false; //-n = no items
			boolean silent = false; //-s = totally silent
			try
			{
				String tok = st.nextToken();
				if ( tok.charAt(0) == '-' )
				{
					if ( tok.indexOf("f") >= 0 ) finalsOnly = true;
					if ( tok.indexOf("s") >= 0 ) silent = true;
					if ( tok.indexOf("n") >= 0 ) noItems = true;
				}
				StringBuffer realArgs = new StringBuffer("");
				while ( st.hasMoreTokens() )
				{
					realArgs.append(st.nextToken());
				}
				arguments = realArgs.toString().trim();
			}
			catch ( Exception e )
			{
				output.println("Unrecognized options. Syntax: PARSE [-fsn] sentence");
			}
			processSentence ( arguments , finalsOnly , silent , noItems );
		}
		public String toString()
		{
			return "PARSE [-fsn] <sentence>";
		}
	}
	
	class LoadGrammarCommand extends Command
	{
		public LoadGrammarCommand()
		{
			super("LOAD");
		}
		public void execute ( String arguments )
		{
			loadGrammarFromFile ( arguments );
		}
		public String toString()
		{
			return "LOAD <path-to-grammar-file>";
		}
	}
	
	class DumpResultsCommand extends Command
	{
		public DumpResultsCommand()
		{
			super("DUMP");
		}
		public void execute ( String arguments )
		{
			StringTokenizer st = new StringTokenizer(arguments," \n\r\t",true);
			boolean finalsOnly = false; //-f = finals only
			boolean noItems = false; //-n = no items
			try
			{
				String tok = st.nextToken();
				if ( tok.charAt(0) == '-' )
				{
					if ( tok.indexOf("f") >= 0 ) finalsOnly = true;
					if ( tok.indexOf("n") >= 0 ) noItems = true;
				}
				StringBuffer realArgs = new StringBuffer("");
				while ( st.hasMoreTokens() )
				{
					realArgs.append(st.nextToken());
				}
				arguments = realArgs.toString().trim();
			}
			catch ( Exception e )
			{
				output.println("Unrecognized options. Syntax: DUMP [-fn] sentence");
			}			
			dumpResults ( arguments , finalsOnly , noItems );
		}
		public String toString()
		{
			return "DUMP [-fn] <path-to-file-to-dump-results-to>";
		}
	}
	
	class ExecuteScriptCommand extends Command
	{
		public ExecuteScriptCommand()
		{
			super("EXEC");
		}
		public void execute ( String arguments )
		{
			executeScript ( arguments );
		}
		public String toString()
		{
			return "EXEC <script-file-to-execute>";
		}
	}
	
	class ExitCommand extends Command
	{
		public ExitCommand()
		{
			super("EXIT");
		}
		public void execute ( String arguments )
		{
			System.exit(0);
		}
		public String toString()
		{
			return "EXIT";
		}
	}
	
	class DefaultCommand extends Command
	{
		public DefaultCommand()
		{
			super(null);
		}
		public void execute ( String arguments )
		{
			output.println("Valid commands: ");
			showCommands();
		}
		public String toString()
		{
			return "Default command";
		}
	}
	
	public void showCommands()
	{
		for ( Iterator it = commands.iterator() ; it.hasNext() ; )
		{
			Command c = (Command) it.next();
			output.println(c);
		}
	}
	
	private PrintStream output = System.out;
	private InputStream input = System.in;
	
	private Grammar grammar;
	private List deductiveSteps;
	private List goalConditions;
	private long timeRef;
	private Agenda agenda = AgendaLoader.getAgenda();
	private DeductiveEngine engine;
	
	private long gramInitTime = 0;
	private long parsingTime = 0;
	
	private int gramInitItems = 0;
	private int gramInitParseItems = 0;
	private int totalItems = 0;
	private int totalParseItems = 0;
	
	private String lastGrammar;
	private String lastSentence;
	
	
	private boolean grammarLoaded;
		
	private void startTimeCount()
	{
		timeRef = System.currentTimeMillis();
	}
	private long getTimeCount()
	{
		return System.currentTimeMillis()-timeRef;
	}
	
	private void dumpResults ( String fileName , boolean finalItemsOption , boolean noItemsOption )
	{
		
		PrintWriter pw;
		
		try
		{
			pw = new PrintWriter ( new OutputStreamWriter ( new FileOutputStream ( new File ( fileName ) , true ) ) );
			pw.println("BEGINNING OF TEST RESULTS");
			pw.println("Test executed at: " + new Date());
			pw.println("Grammar: " + lastGrammar);
			pw.println("Sentence: " + lastSentence);
			pw.println("Grammar init time (ms): " + gramInitTime);
			pw.println("Parsing time (ms): " + parsingTime);
			pw.println("Grammar init items: " + gramInitItems);
			pw.println("Total items: " + totalItems);
			pw.println("Items generated during parsing time: " + (totalItems-gramInitItems));
			pw.println("Grammar init parse items: " + gramInitParseItems);
			pw.println("Total parse items: " + totalParseItems);
			pw.println("Parse items generated during parsing time: " + (totalParseItems-gramInitParseItems));
			pw.println("Item list:");
			List lItems = ItemHandler.instance().getAll();
			if ( !noItemsOption )
			{
					for ( int i = 0 ; i < lItems.size() ; i++ )
					{
						Object[] item = (Object[]) lItems.get(i);
						
						if ( !finalItemsOption )
							pw.println(Items.toString(item));
						else
							for ( int j = 0 ; j < goalConditions.size() ; j++ )
							{
								List result = ((DeductiveStep)goalConditions.get(j)).applyTo(item);
								if ( result != null && result.size() > 0 )
								{
									pw.println(Items.toString(item));
									break; //pasamos al siguiente item, no queremos imprimir ste varias veces aunque cumpla varias condiciones
								}
							}
					}
			}
			pw.println("End of item list.");
			pw.println("END OF TEST RESULTS");
			pw.close();
			output.println("Results dumped to file " + fileName);
		}
		catch ( Throwable t )
		{
			System.err.println("Error writing dump to file " + fileName + ": ");
			t.printStackTrace();
		}
		
	}
	
	private void executeScript ( String fileName )
	{
		
		BufferedReader br;
		String linea = " ";
		
		try
		{
			br = new BufferedReader ( new InputStreamReader ( new FileInputStream ( new File ( fileName ) ) ) );
			while(linea != null)
			{
			    linea = br.readLine();
				processInput(linea);
			}
		}
		catch ( Throwable t )
		{
			System.err.println("Error reading script from file " + fileName + ": ");
			t.printStackTrace();
		}
		
	}
	
	/**
	 * Parses all sentences in a file.
	 * @param fileName The file to read sentences from.
	 */
	private void parseFile ( String fileName , String resultFileName )
	{
		try
		{
			File f = new File ( fileName );
			//String resultFileName = f.getName()+".result";
			
			BufferedReader br = new BufferedReader ( new InputStreamReader ( new FileInputStream ( f ) ) );
			String line;
			while ( (line = br.readLine() ) != null )
			{
				if ( line.startsWith(";") ) ;
				else
				{
					processSentence ( line , false , false , true );
					dumpResults ( resultFileName , true , false );
				}
			}
			System.out.println("Parsing of file " + fileName + " done.");
			
		}
		catch ( Throwable t )
		{
			System.err.println("Error parsing sentences from file " + fileName + ": ");
			t.printStackTrace();
		}
		
	}
	
	private void loadGrammarFromFile ( String fileName )
	{
		
		lastGrammar = fileName;
		
		output.println("Preparando algoritmo para gramtica...");
		
		try
		{
			grammar = GrammarLoader.getGrammar ( new File(fileName) );
		}
		catch ( Throwable t )
		{
			System.err.println("Error reading grammar from file " + fileName + ": ");
			t.printStackTrace();
		}
		
		//instanciar los pasos deductivos y StepHandler
		startTimeCount();
		
		deductiveSteps = StepFactory.instanceSteps(grammar);
		goalConditions = StepFactory.instanceGoalConditions(grammar);
		new ItemHandler();
		new StepHandler();
		int sz = deductiveSteps.size();
		for ( int i = 0 ; i < sz ; i++ )
		{
			StepHandler.instance().addStep((DeductiveStep)deductiveSteps.get(i));
		}
		engine = DeductiveEngineLoader.getEngine( agenda , deductiveSteps , goalConditions , ItemHandler.instance() , StepHandler.instance() );
		
		long time = getTimeCount();
		output.println(deductiveSteps.size() + " pasos deductivos instanciados en " + time + " ms");
	
		
		//ejecutar la mquina deductiva para hacer todo aquello que se pueda hacer sin la frase
		startTimeCount();
		engine.preprocess();
		time = getTimeCount();
		output.println(ItemHandler.instance().getAll().size() + " items computados en " + time + " ms"); 
		
		ItemHandler.fastSave();
		
		output.println("Gramtica preparada.");
		gramInitTime = time;
		gramInitItems = ItemHandler.instance().getAll().size();
		gramInitParseItems = 0;
		List itList = ItemHandler.instance().getAll();
		for ( int i = 0 ; i < itList.size() ; i++ )
		{
			if ( Items.toString((Object[])itList.get(i)).contains("Parse") ) gramInitParseItems++;
		}
		
		grammarLoaded = true;
		gramInitTime = time;
		
	}
	
	private void processSentence ( String sentence , boolean goalItemsOption , boolean silentOption , boolean noItemsOption )
	{	
		
		lastSentence = sentence;
		
		StringTokenizer st = new StringTokenizer( sentence );
		//leer terminales
		List tokens = new ArrayList(); //of items
		int sentencelen = 0;
		//apply k-function to generate items from sentence!!
		while ( st.hasMoreTokens() )
		{
			String tokText = st.nextToken().trim();
			Symbol tokSym = SymbolTable.instance().symbolFor(tokText.replace(':','%').trim());
			Element[] item = new Element[3];
			item[0] = tokSym;
			item[1] = StringPosition.groundFromValue(sentencelen);
			sentencelen++;
			item[2] = StringPosition.groundFromValue(sentencelen);
			tokens.add(item);
		}
		StringPosition.setMaxValue(sentencelen);
		//new ItemHandler();
		startTimeCount();
		ItemHandler.fastRestore();
		long time = getTimeCount();
		if ( !silentOption ) output.println("ItemHandler restaurada en " + time + " ms");
		for ( int i = 0 ; i < tokens.size() ; i++ )
		{
			ItemHandler.instance().addItem((Element[])tokens.get(i));
			agenda.add ( (Element[]) tokens.get(i) );
		}
		
		//resultArea.clear();
		engine.setItemHandler(ItemHandler.instance()); //this is due to fastRestore() calling new ItemHandler() and changing the registered instance
		startTimeCount();
		engine.process();
		time = getTimeCount();
		if ( !silentOption ) output.println(ItemHandler.instance().getAll().size() + " items computados en " + time + " ms"); 
		
		parsingTime = time;
		totalItems = ItemHandler.instance().getAll().size();
		
		totalParseItems = 0;
		List itList = ItemHandler.instance().getAll();
		for ( int i = 0 ; i < itList.size() ; i++ )
		{
			if ( Items.toString((Object[])itList.get(i)).contains("Parse") ) totalParseItems++;
		}
		
		//sacar items por el rea de items
		List lItems = ItemHandler.instance().getAll();
		
		if ( !goalItemsOption )
		{
			if ( !silentOption ) output.println("Items obtenidos:");
			
			StringBuffer accum = new StringBuffer("");
			
			for ( int i = 0 ; i < lItems.size() ; i++ )
			{
				Object[] item = (Object[]) lItems.get(i);
			
				accum.append("\n"+Items.toString(item));
				//resultArea.println(Items.toString(item));
				if ( i % 1000 == 999 )
				{
					if ( !silentOption && !noItemsOption ) output.println(accum.toString());
					accum = new StringBuffer("");
				}
			}
			if ( !silentOption && !noItemsOption ) output.println(accum.toString());
			if ( !silentOption ) output.println("Fin de lista de items.");
		}
		else
		{
			if ( !silentOption ) output.println("Items finales:");
			List goalConditions = StepFactory.instanceGoalConditions(grammar);
			boolean found = false;
			for ( int i = 0 ; i < lItems.size() ; i++ )
			{
				Object[] item = (Object[]) lItems.get(i);
				if ( goalConditions == null || goalConditions.size() == 0 )
					output.println(Items.toString(item));
				else
				{
					for ( int j = 0 ; j < goalConditions.size() ; j++ )
					{
						List result = ((DeductiveStep)goalConditions.get(j)).applyTo(item);
						if ( result != null && result.size() > 0 )
						{
							found = true;
							if ( !silentOption && !noItemsOption ) output.println(Items.toString(item));
							break; //pasamos al siguiente item, no queremos imprimir ste varias veces aunque cumpla varias condiciones
						}
					}
				}
			}
			if ( !found )
				if ( !silentOption ) output.println("No hay items finales.");
		}
		
	}
		
	private LinkedList commands = new LinkedList();
	private Command defaultCommand;
	
	public void addCommand ( Command cmd )
	{
		commands.add(cmd);
	}
	
	public void processInput ( String input )
	{
		StringTokenizer st = new StringTokenizer(input);
		try
		{
			String firstWord = st.nextToken();
			for ( Iterator it = commands.iterator() ; it.hasNext() ; )
			{
				Command com = (Command) it.next();
				if ( firstWord.equalsIgnoreCase(com.getCommandName()) )
				{
					String rest = input.substring(com.getCommandName().length());
					String args = rest.trim();
					com.execute(args);
					return;
				}
			}
			defaultCommand.execute(input);
		}
		catch ( Exception e )
		{
			output.println("Empty line issued.");
			defaultCommand.execute(input);
		}
	}

	public ParserConsole(String[] args) throws IOException
	{
		addCommand ( new LoadGrammarCommand() );
		addCommand ( new ParseSentenceCommand() );
		addCommand ( new ExecuteScriptCommand() );
		addCommand ( new DumpResultsCommand() );
		addCommand ( new ParseFileCommand() );
		addCommand ( new ExitCommand() );
		defaultCommand = new DefaultCommand();
		BufferedReader br = new BufferedReader ( new InputStreamReader ( System.in ) );
		
		if ( args != null && args.length > 0 && args[0] != null )
		{
			//exec script
			executeScript(args[0]);
		}
		
		for(;;)
		{
			output.println(">");
			String str = br.readLine();
			processInput(str);
		}
		
	}
	
	public static void main(String[] args) throws IOException
	{
		new ParserConsole(args);
	}
}
