package agenda;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * Superclass of the implementations of Agenda as a priority queue.
 * A priority queue agenda defines the "first" item in the agenda as the first
 * that was inserted into it among the items with highest priority.
 * The way to define priority and obtain the priority of an item is left
 * to the concrete implementations of this class.
 * However, PriorityQueueAgenda requires the priority of items to be
 * represented as a nonnegative int, with 0 representing the highest priority,
 * and greater numbers representing lower priorities. Also, this implementation is
 * designed to be fast and efficient when only the first few nonnegative ints are
 * used for priorities (e.g. 0 to 10, o to 32...) but would be extremely inefficient
 * if larger ints were used. In the latter case, this implementation must not be used
 * at all.
 * @author carlos
 *
 * Created at regulus, 11-nov-2006 12:36:18
 */
public abstract class PriorityQueueAgenda implements Agenda
{
	
	/**
	 * List of simple queues, one for each priority.
	 */
	private ArrayList queues = new ArrayList();
	
	/**
	 * Stores the index of the queue holding the first item, for fast access.
	 */
	private int currentQueue = 0;
	
	/**
	 * Stores the total size.
	 */
	private int size = 0;
	
	/**
	 * This method, the only one that need be overridden by subclasses,
	 * returns the priority of a given item as a nonnegative int.
	 * @param item The item whose priority must be obtained.
	 * @return A small, nonnegative int representing the item's priority.
	 */
	protected abstract int getPriority ( Object item );
	
	public boolean add ( Object item )
	{
		
		//obtain the item's priority
		int priority = getPriority ( item );
		
		//if not enough queues exist in the list, add them
		while ( queues.size() <= priority )
		{
			LinkedList newQueue = new LinkedList();
			queues.add(newQueue);
		}
		
		//update the pointer to the current queue if necessary
		if ( isEmpty() || priority < currentQueue ) currentQueue = priority;
		
		//add the element to its corresponding queue
		((LinkedList)queues.get(priority)).add(item);
		
		//increment the size
		size++;
		
		return true;
		
	}
	
	public boolean isEmpty()
	{
		//return ( ((LinkedList)queues.get(currentQueue)).isEmpty() );
		return ( size == 0 );
	}
	
	public Object removeFirst ( ) throws NoSuchElementException
	{
		
		LinkedList ourQueue = (LinkedList)queues.get(currentQueue);
		
		//this line throws NoSuchElementException when appropriate
		Object returnValue = ourQueue.removeFirst();
		
		//{if we reach this point, returnValue should be a valid item to return}
		
		if ( ourQueue.isEmpty() )
		{
			//upgrade the current queue pointer
			while ( currentQueue < queues.size()-1 && ((LinkedList)queues.get(currentQueue)).isEmpty() )
				currentQueue++;
		}
		
		//decrement the size
		size--;
		
		return returnValue;
		
	}
	
	public int size()
	{
		return size;
	}
	
	public void clear()
	{
		queues.clear();
		currentQueue = 0;
		size = 0;
	}
	
	
	
}